'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var tslib_1 = require('tslib');
var debeSocket = require('debe-socket');
var debeDelta = require('debe-delta');
var debe = require('debe');
var debeAdapter = require('debe-adapter');

var CHANNELS = {
    SEND: 'sendChanges',
    SEND_DELTA: 'sendDeltaChanges',
    COUNT_INITIAL: 'countInitialChanges',
    COUNT_INITIAL_DELTA: 'countInitialDeltaChanges',
    FETCH_INITIAL: 'fetchInitialChanges',
    FETCH_INITIAL_DELTA: 'fetchInitialDeltaChanges',
    FETCH_MISSING_DELTA: 'fetchMissingDelta',
    SUBSCRIBE_CHANGES: "subscription",
    SUBSCRIBE_DELTA_CHANGES: "subscribeDelta"
};

var syncstateTable = 'syncstate';
var batchSize = 1000;

function hashCode(str) {
    if (str === void 0) { str = ''; }
    var hash = 0, i, chr;
    if (str.length === 0)
        return hash;
    for (i = 0; i < str.length; i++) {
        chr = str.charCodeAt(i);
        hash = (hash << 5) - hash + chr;
        hash |= 0;
    }
    return hash;
}
function getLastItemRev(items, revField, comparer) {
    if (items === void 0) { items = []; }
    var result = items[items.length - 1]
        ? items[items.length - 1][revField]
        : undefined;
    if (!comparer || result > comparer) {
        return result;
    }
    return comparer;
}
var maxRetries = 5;
var maxTimeout = 2500;
var defaultInterval = function (tries) {
    switch (tries) {
        case 0:
            return 100;
        case 1:
            return 200;
        case 2:
            return 400;
        case 3:
            return 800;
        case 4:
            return 1600;
        default:
            return 3200;
    }
};
function waitFor(cb, _a) {
    var _b = (_a === void 0 ? {} : _a).interval, interval = _b === void 0 ? defaultInterval : _b;
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        var tries, result;
        return tslib_1.__generator(this, function (_c) {
            switch (_c.label) {
                case 0:
                    tries = 0;
                    _c.label = 1;
                case 1: return [4, cb()];
                case 2:
                    if (!(!(result = _c.sent()) && tries <= maxRetries)) return [3, 4];
                    tries += 1;
                    return [4, new Promise(function (yay) {
                            return setTimeout(yay, typeof interval === 'function' ? interval(tries) : interval);
                        })];
                case 3:
                    _c.sent();
                    return [3, 1];
                case 4: return [2, Promise.resolve(result)];
            }
        });
    });
}
function getBigger(comparer0, comparer1) {
    if (!comparer0 || !comparer1) {
        return undefined;
    }
    if (!comparer1 || comparer0 > comparer1) {
        return comparer0;
    }
    return comparer1;
}
function batchTransfer(_a) {
    var fetchCount = _a.fetchCount, fetchItems = _a.fetchItems, transferItems = _a.transferItems, reporter = _a.reporter, shouldCancel = _a.shouldCancel;
    return tslib_1.__awaiter(this, void 0, void 0, function () {
        var canceled, changeCount, current, page, items, transferred, changes, _b, _c;
        return tslib_1.__generator(this, function (_d) {
            switch (_d.label) {
                case 0:
                    canceled = shouldCancel ? shouldCancel : function () { return false; };
                    return [4, fetchCount()];
                case 1:
                    changeCount = _d.sent();
                    current = changeCount;
                    page = 0;
                    items = [];
                    transferred = [];
                    _d.label = 2;
                case 2:
                    if (!(current > 0)) return [3, 8];
                    return [4, fetchItems(page)];
                case 3:
                    changes = _d.sent();
                    if (!(changes.length && !canceled())) return [3, 6];
                    if (!transferItems) return [3, 5];
                    _c = (_b = transferred).concat;
                    return [4, transferItems(changes)];
                case 4:
                    transferred = _c.apply(_b, [_d.sent()]);
                    _d.label = 5;
                case 5:
                    items = items.concat(changes);
                    return [3, 7];
                case 6: return [3, 8];
                case 7:
                    current = current - changes.length;
                    page = page + 1;
                    if (reporter) {
                        reporter(Math.abs((100 / changeCount) * current - 100));
                    }
                    return [3, 2];
                case 8:
                    if (transferItems) {
                        return [2, [items, transferred]];
                    }
                    return [2, [items, []]];
            }
        });
    });
}

var SyncState = (function () {
    function SyncState(db) {
        var _this = this;
        this.state = {};
        this.local = function (collection) {
            if (!_this.state[collection]) {
                return undefined;
            }
            return _this.state[collection].local;
        };
        this.remote = function (collection) {
            if (!_this.state[collection]) {
                return undefined;
            }
            return _this.state[collection].remote;
        };
        this.promise = Promise.resolve();
        this.init = function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var state;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, this.db.get(syncstateTable)];
                    case 1:
                        state = _a.sent();
                        if (state) {
                            this.state = state;
                        }
                        return [2];
                }
            });
        }); };
        this.update = function (collection, local, remote) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                if (!this.state[collection]) {
                    this.state[collection] = {};
                }
                if (local) {
                    this.state[collection].local = local;
                }
                if (remote) {
                    this.state[collection].remote = remote;
                }
                this.promise = this.promise.then(function () {
                    return _this.db.insert(syncstateTable, tslib_1.__assign({}, _this.state));
                });
                return [2];
            });
        }); };
        this.db = db;
    }
    return SyncState;
}());

var SEND_DELTA = CHANNELS.SEND_DELTA, FETCH_MISSING_DELTA = CHANNELS.FETCH_MISSING_DELTA;
var delta = {
    initialDown: function (collection, since, count, sync) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var _a, remoteChanges;
            var _this = this;
            return tslib_1.__generator(this, function (_b) {
                switch (_b.label) {
                    case 0: return [4, batchTransfer({
                            fetchCount: function () { return count; },
                            fetchItems: function (page) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                return tslib_1.__generator(this, function (_a) {
                                    return [2, sync.socket.invoke(CHANNELS.FETCH_INITIAL_DELTA, {
                                            type: collection,
                                            since: since,
                                            page: page
                                        })];
                                });
                            }); },
                            shouldCancel: function () { return sync.isClosing; }
                        })];
                    case 1:
                        _a = tslib_1.__read.apply(void 0, [_b.sent(), 1]), remoteChanges = _a[0];
                        if (remoteChanges.length && !sync.isClosing) {
                            return [2, function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                    var options, o;
                                    return tslib_1.__generator(this, function (_a) {
                                        switch (_a.label) {
                                            case 0:
                                                options = {
                                                    synced: 'client'
                                                };
                                                return [4, debeDelta.merge(sync.db, collection, remoteChanges, options).catch(function (x) { return console.error('Error while client.insert', x); })];
                                            case 1:
                                                _a.sent();
                                                o = remoteChanges[remoteChanges.length - 1];
                                                return [2, [undefined, o && o[2]]];
                                        }
                                    });
                                }); }];
                        }
                        return [2, function () { return tslib_1.__awaiter(_this, void 0, void 0, function () { return tslib_1.__generator(this, function (_a) {
                                return [2, [undefined, undefined]];
                            }); }); }];
                }
            });
        });
    },
    initialUp: function (collection, since, count, sync) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var latestResult, latestUpload, err_1;
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        latestResult = undefined;
                        latestUpload = undefined;
                        _a.label = 1;
                    case 1:
                        _a.trys.push([1, 3, , 4]);
                        return [4, batchTransfer({
                                fetchCount: function () { return count; },
                                fetchItems: function (page) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                    return tslib_1.__generator(this, function (_a) {
                                        return [2, sync.db
                                                .all(collection, {
                                                where: since ? ['rev > ?', since] : undefined,
                                                orderBy: ['rev ASC', 'id ASC'],
                                                limit: batchSize,
                                                offset: page * batchSize
                                            })
                                                .then(function (x) {
                                                if (x.length) {
                                                    latestUpload = x[x.length - 1].rev;
                                                }
                                                return x;
                                            })];
                                    });
                                }); },
                                transferItems: function (items) {
                                    return delta
                                        .up(collection, items, {
                                        delta: items.map(function (x) { return [x.id, x['merge'], x.rev]; })
                                    }, sync)
                                        .then(function (x) {
                                        latestResult = x;
                                        return x;
                                    });
                                },
                                shouldCancel: function () { return sync.isClosing; }
                            })];
                    case 2:
                        _a.sent();
                        return [2, [latestUpload, latestResult ? latestResult.latestRev : undefined]];
                    case 3:
                        err_1 = _a.sent();
                        return [2, [undefined, undefined]];
                    case 4: return [2];
                }
            });
        });
    },
    listen: function (sync) {
        var _this = this;
        var channel = sync.socket.subscribe(CHANNELS.SUBSCRIBE_DELTA_CHANGES);
        (function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var channel_1, channel_1_1, data, release, _a, type, items, _b, options, newItems, result, o0, o, e_1_1;
            var e_1, _c;
            return tslib_1.__generator(this, function (_d) {
                switch (_d.label) {
                    case 0:
                        _d.trys.push([0, 10, 11, 16]);
                        channel_1 = tslib_1.__asyncValues(channel);
                        _d.label = 1;
                    case 1: return [4, channel_1.next()];
                    case 2:
                        if (!(channel_1_1 = _d.sent(), !channel_1_1.done)) return [3, 9];
                        data = channel_1_1.value;
                        if (sync.isClosing) {
                            return [2];
                        }
                        return [4, sync.initialSyncComplete];
                    case 3:
                        _d.sent();
                        release = sync.registerTask();
                        _a = tslib_1.__read(data, 3), type = _a[0], items = _a[1], _b = _a[2], options = _b === void 0 ? {} : _b;
                        options.synced = sync.socket.id;
                        return [4, debeDelta.merge(sync.db, type, items, options).catch(function (x) { return console.error('Error while client.insert', x); })];
                    case 4:
                        newItems = _d.sent();
                        if (!options.unsucessful.length) return [3, 7];
                        return [4, sync.socket.invoke(FETCH_MISSING_DELTA, [
                                type,
                                options.unsucessful
                            ])];
                    case 5:
                        result = _d.sent();
                        return [4, debeDelta.merge(sync.db, type, result, options).catch(function (x) { return console.error('Error while client.insert', x); })];
                    case 6:
                        newItems = _d.sent();
                        _d.label = 7;
                    case 7:
                        o0 = newItems[newItems.length - 1];
                        o = items[items.length - 1];
                        sync.syncState.update(type, o0 ? o0.rev : undefined, o && o[2]);
                        release();
                        _d.label = 8;
                    case 8: return [3, 1];
                    case 9: return [3, 16];
                    case 10:
                        e_1_1 = _d.sent();
                        e_1 = { error: e_1_1 };
                        return [3, 16];
                    case 11:
                        _d.trys.push([11, , 14, 15]);
                        if (!(channel_1_1 && !channel_1_1.done && (_c = channel_1.return))) return [3, 13];
                        return [4, _c.call(channel_1)];
                    case 12:
                        _d.sent();
                        _d.label = 13;
                    case 13: return [3, 15];
                    case 14:
                        if (e_1) throw e_1.error;
                        return [7];
                    case 15: return [7];
                    case 16: return [2];
                }
            });
        }); })();
        return {
            wait: new Promise(function (yay) { return setTimeout(yay, 10); })
        };
    },
    up: function (collection, items, options, sync) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var _a, _b, failed, latestRev, missing;
            return tslib_1.__generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        if (sync.isClosing) {
                            return [2];
                        }
                        return [4, sync.socket.invoke(SEND_DELTA, [collection, options['delta']])];
                    case 1:
                        _a = _c.sent(), _b = _a.failed, failed = _b === void 0 ? [] : _b, latestRev = _a.latestRev;
                        if (sync.isClosing) {
                            return [2];
                        }
                        if (!failed.length) return [3, 4];
                        return [4, sync.db.all(collection, {
                                id: failed,
                                select: ['id', 'merge']
                            })];
                    case 2:
                        missing = _c.sent();
                        if (sync.isClosing) {
                            return [2];
                        }
                        return [4, sync.socket
                                .invoke(SEND_DELTA, [collection, missing.map(function (x) { return [x.id, x['merge']]; })])
                                .then(function (x) { return (x.latestRev > latestRev ? x.latestRev : latestRev); })];
                    case 3:
                        latestRev = _c.sent();
                        _c.label = 4;
                    case 4: return [2, latestRev];
                }
            });
        });
    }
};

var basic = {
    initialDown: function (collection, since, count, sync) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var _a, remoteChanges;
            var _this = this;
            return tslib_1.__generator(this, function (_b) {
                switch (_b.label) {
                    case 0: return [4, batchTransfer({
                            fetchCount: function () { return count; },
                            fetchItems: function (page) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                return tslib_1.__generator(this, function (_a) {
                                    return [2, sync.socket.invoke(CHANNELS.FETCH_INITIAL, {
                                            type: collection,
                                            since: since,
                                            page: page
                                        })];
                                });
                            }); },
                            shouldCancel: function () { return sync.isClosing; }
                        })];
                    case 1:
                        _a = tslib_1.__read.apply(void 0, [_b.sent(), 1]), remoteChanges = _a[0];
                        return [2, function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                var options, o;
                                return tslib_1.__generator(this, function (_a) {
                                    switch (_a.label) {
                                        case 0:
                                            if (!(remoteChanges.length && !sync.isClosing)) return [3, 2];
                                            options = {
                                                synced: 'client'
                                            };
                                            return [4, sync.db.insert(collection, remoteChanges, options)];
                                        case 1:
                                            _a.sent();
                                            o = remoteChanges[remoteChanges.length - 1];
                                            return [2, [undefined, o && o[2]]];
                                        case 2: return [2, [undefined, undefined]];
                                    }
                                });
                            }); }];
                }
            });
        });
    },
    initialUp: function (collection, since, count, sync) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var latestResult, latestUpload, err_1;
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        latestResult = undefined;
                        latestUpload = undefined;
                        _a.label = 1;
                    case 1:
                        _a.trys.push([1, 3, , 4]);
                        return [4, batchTransfer({
                                fetchCount: function () { return count; },
                                fetchItems: function (page) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                    return tslib_1.__generator(this, function (_a) {
                                        return [2, sync.db
                                                .all(collection, {
                                                where: since ? ['rev > ?', since] : undefined,
                                                orderBy: ['rev ASC', 'id ASC'],
                                                limit: batchSize,
                                                offset: page * batchSize
                                            })
                                                .then(function (x) {
                                                if (x.length) {
                                                    latestUpload = x[x.length - 1].rev;
                                                }
                                                return x;
                                            })];
                                    });
                                }); },
                                transferItems: function (items) { return basic.up(collection, items, {}, sync); },
                                shouldCancel: function () { return sync.isClosing; }
                            })];
                    case 2:
                        _a.sent();
                        return [2, [latestUpload, latestResult ? latestResult.latestRev : undefined]];
                    case 3:
                        err_1 = _a.sent();
                        return [2, [undefined, undefined]];
                    case 4: return [2];
                }
            });
        });
    },
    listen: function (sync) {
        var _this = this;
        var channel = sync.socket.subscribe(CHANNELS.SUBSCRIBE_CHANGES);
        (function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var channel_1, channel_1_1, data, _a, type, items, _b, options, release, lastRemoteRev, newItems, lastLocalRev, e_1_1;
            var e_1, _c;
            return tslib_1.__generator(this, function (_d) {
                switch (_d.label) {
                    case 0:
                        _d.trys.push([0, 7, 8, 13]);
                        channel_1 = tslib_1.__asyncValues(channel);
                        _d.label = 1;
                    case 1: return [4, channel_1.next()];
                    case 2:
                        if (!(channel_1_1 = _d.sent(), !channel_1_1.done)) return [3, 6];
                        data = channel_1_1.value;
                        if (sync.isClosing) {
                            return [2];
                        }
                        return [4, sync.initialSyncComplete];
                    case 3:
                        _d.sent();
                        _a = tslib_1.__read(data, 3), type = _a[0], items = _a[1], _b = _a[2], options = _b === void 0 ? {} : _b;
                        release = sync.registerTask();
                        options.synced = sync.socket.id;
                        lastRemoteRev = items[items.length - 1] && items[items.length - 1].rev;
                        return [4, sync.db.insert(type, items, options)];
                    case 4:
                        newItems = _d.sent();
                        lastLocalRev = newItems[newItems.length - 1] && newItems[newItems.length - 1].rev;
                        sync.syncState.update(type, lastLocalRev, lastRemoteRev);
                        release();
                        _d.label = 5;
                    case 5: return [3, 1];
                    case 6: return [3, 13];
                    case 7:
                        e_1_1 = _d.sent();
                        e_1 = { error: e_1_1 };
                        return [3, 13];
                    case 8:
                        _d.trys.push([8, , 11, 12]);
                        if (!(channel_1_1 && !channel_1_1.done && (_c = channel_1.return))) return [3, 10];
                        return [4, _c.call(channel_1)];
                    case 9:
                        _d.sent();
                        _d.label = 10;
                    case 10: return [3, 12];
                    case 11:
                        if (e_1) throw e_1.error;
                        return [7];
                    case 12: return [7];
                    case 13: return [2];
                }
            });
        }); })();
        return {
            wait: new Promise(function (yay) { return setTimeout(yay, 10); })
        };
    },
    up: function (collection, items, options, sync) {
        if (sync.isClosing) {
            return Promise.resolve();
        }
        return sync.socket
            .invoke(CHANNELS.SEND, [collection, items])
            .catch(function (err) {
            if (!sync.isClosing) {
                throw err;
            }
        });
    }
};

var SyncEngine = (function () {
    function SyncEngine(socket, db, state) {
        this.initialSyncComplete = Promise.resolve();
        this.cancels = [];
        this.tasks = {};
        this.isClosing = false;
        this.socket = socket;
        this.db = db;
        this.syncState = state;
    }
    Object.defineProperty(SyncEngine.prototype, "isInitialComplete", {
        get: function () {
            return this.tasks['initial'] !== true;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SyncEngine.prototype, "isSyncing", {
        get: function () {
            return Object.keys(this.tasks).length > 0;
        },
        enumerable: true,
        configurable: true
    });
    SyncEngine.prototype.initial = function (collections, isDelta) {
        var release = this.registerTask('initial');
        this.initialSyncComplete = this.initialSync(collections, isDelta).finally(function () {
            release();
        });
        this.cancels.push(this.listenToDatabase(collections, isDelta));
        this.listenToSync();
    };
    SyncEngine.prototype.close = function () {
        var _this = this;
        this.isClosing = true;
        this['_close'] = function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, Promise.all(this.cancels.map(function (x) { return x(); }))];
                    case 1:
                        _a.sent();
                        return [2];
                }
            });
        }); };
    };
    SyncEngine.prototype.listenToSync = function () {
        var cancels = [];
        cancels.push(delta.listen(this));
        cancels.push(basic.listen(this));
    };
    SyncEngine.prototype.getRemoteChangeCount = function (socket, collections, syncState) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var remoteCounts;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        remoteCounts = {};
                        return [4, Promise.all(collections.map(function (key) {
                                return socket
                                    .invoke(CHANNELS.COUNT_INITIAL_DELTA, {
                                    type: key,
                                    since: syncState.remote(key)
                                })
                                    .then(function (result) { return (remoteCounts[key] = result); });
                            }))];
                    case 1:
                        _a.sent();
                        return [2, remoteCounts];
                }
            });
        });
    };
    SyncEngine.prototype.getLocalChangeCount = function (db, collections, syncState) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var localCounts;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        localCounts = {};
                        return [4, Promise.all(collections.map(function (key) {
                                var begin = syncState.local(key);
                                return db
                                    .count(key, {
                                    where: begin ? ['rev > ?', begin] : undefined
                                })
                                    .then(function (result) { return (localCounts[key] = result); });
                            }))];
                    case 1:
                        _a.sent();
                        return [2, localCounts];
                }
            });
        });
    };
    SyncEngine.prototype.getTaskKey = function () {
        var key = new Date().getTime();
        var count = 0;
        while (this.tasks[key + "-" + count] !== undefined) {
            count = count + 1;
        }
        return key + "-" + count;
    };
    SyncEngine.prototype.registerTask = function (initialKey) {
        var _this = this;
        var key = initialKey || this.getTaskKey();
        this.tasks[key] = true;
        return function () {
            delete _this.tasks[key];
        };
    };
    SyncEngine.prototype.listenToDatabase = function (collections, isDelta) {
        var _this = this;
        return this.db.listen('*', function (error, items, options, key) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var release;
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (error) {
                            throw error;
                        }
                        return [4, this.initialSyncComplete];
                    case 1:
                        _a.sent();
                        if (this.socket.state !== 'open') {
                            return [2];
                        }
                        release = this.registerTask();
                        if (!key ||
                            collections.indexOf(key) === -1 ||
                            options.synced === this.socket.id) {
                            return [2];
                        }
                        if (isDelta(key)) {
                            delta.up(key, items, options, this);
                        }
                        else {
                            basic
                                .up(key, items, options, this)
                                .catch(function (err) { return console.log('ERR', err, err.stack, _this.socket.state); })
                                .finally(function () {
                                release();
                            });
                        }
                        return [2];
                }
            });
        }); });
    };
    SyncEngine.prototype.initialSync = function (collections, isDelta) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var remoteCounts, localCounts, collections_1, collections_1_1, key, down, persist, up, _a, local2, remote2, _b, local, remote, e_1_1;
            var e_1, _c;
            return tslib_1.__generator(this, function (_d) {
                switch (_d.label) {
                    case 0: return [4, this.getRemoteChangeCount(this.socket, collections, this.syncState)];
                    case 1:
                        remoteCounts = _d.sent();
                        return [4, this.getLocalChangeCount(this.db, collections, this.syncState)];
                    case 2:
                        localCounts = _d.sent();
                        _d.label = 3;
                    case 3:
                        _d.trys.push([3, 10, 11, 12]);
                        collections_1 = tslib_1.__values(collections), collections_1_1 = collections_1.next();
                        _d.label = 4;
                    case 4:
                        if (!!collections_1_1.done) return [3, 9];
                        key = collections_1_1.value;
                        down = isDelta(key) ? delta.initialDown : basic.initialDown;
                        return [4, down(key, this.syncState.remote(key), remoteCounts[key], this)];
                    case 5:
                        persist = _d.sent();
                        up = isDelta(key) ? delta.initialUp : basic.initialUp;
                        return [4, up(key, this.syncState.local(key), localCounts[key], this)];
                    case 6:
                        _a = tslib_1.__read.apply(void 0, [_d.sent(), 2]), local2 = _a[0], remote2 = _a[1];
                        return [4, persist()];
                    case 7:
                        _b = tslib_1.__read.apply(void 0, [_d.sent(), 2]), local = _b[0], remote = _b[1];
                        this.syncState.update(key, local, remote);
                        this.syncState.update(key, local2, remote2);
                        _d.label = 8;
                    case 8:
                        collections_1_1 = collections_1.next();
                        return [3, 4];
                    case 9: return [3, 12];
                    case 10:
                        e_1_1 = _d.sent();
                        e_1 = { error: e_1_1 };
                        return [3, 12];
                    case 11:
                        try {
                            if (collections_1_1 && !collections_1_1.done && (_c = collections_1.return)) _c.call(collections_1);
                        }
                        finally { if (e_1) throw e_1.error; }
                        return [7];
                    case 12: return [2];
                }
            });
        });
    };
    return SyncEngine;
}());

function ensureSync(db) {
    var backend = db.dispatcher;
    if (backend.middlewares) {
        debeAdapter.addMiddleware(backend, {
            name: 'sync',
            collection: function (collection) {
                if (collection['sync'] && collection['sync'] === 'delta') {
                    debeAdapter.addPlugin(collection, 'delta');
                }
                return collection;
            },
            collections: function (collections) {
                collections[syncstateTable] = debe.ensureCollection({
                    name: syncstateTable,
                    fields: {
                        state: 'json'
                    }
                });
                return collections;
            }
        });
        debeDelta.deltaPlugin(backend);
    }
}

var Sync = (function () {
    function Sync(db, _a) {
        var _b = tslib_1.__read(_a === void 0 ? [] : _a, 3), _c = _b[0], hostname = _c === void 0 ? 'localhost' : _c, _d = _b[1], port = _d === void 0 ? 8000 : _d, _e = _b[2], secure = _e === void 0 ? undefined : _e;
        this.db = db;
        this.syncState = new SyncState(db);
        ensureSync(this.db);
        this.db['sync'] = this;
        this.socket = debeSocket.create({
            secure: secure === true || secure === false ? secure : port === 443,
            hostname: hostname,
            port: port
        });
    }
    Sync.prototype.initialize = function () {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, this.db.initialize()];
                    case 1:
                        _a.sent();
                        this.loop();
                        return [2, this];
                }
            });
        });
    };
    Object.defineProperty(Sync.prototype, "initialSync", {
        get: function () {
            return this.engine.initialSyncComplete;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Sync.prototype, "isSyncing", {
        get: function () {
            return this.engine.isSyncing;
        },
        enumerable: true,
        configurable: true
    });
    Sync.prototype.close = function () {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!this['_close']) return [3, 2];
                        return [4, this['_close']()];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2:
                        this.socket.disconnect();
                        return [4, this.db.close()];
                    case 3:
                        _a.sent();
                        return [2];
                }
            });
        });
    };
    Sync.prototype.disconnect = function () {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!this['_close']) return [3, 2];
                        return [4, this['_close']()];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2:
                        this.socket.disconnect();
                        return [2];
                }
            });
        });
    };
    Sync.prototype.connect = function () {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var collections, isDelta, keys;
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, this.db.initialize()];
                    case 1:
                        _a.sent();
                        return [4, this.syncState.init()];
                    case 2:
                        _a.sent();
                        this.engine = new SyncEngine(this.socket, this.db, this.syncState);
                        collections = this.db.dispatcher.collections;
                        isDelta = function (key) { return collections[key]['sync'] === 'delta'; };
                        keys = Object.keys(collections).filter(function (x) { return collections[x]['sync']; });
                        this.engine.initial(keys, isDelta);
                        this['_close'] = function () { return tslib_1.__awaiter(_this, void 0, void 0, function () { return tslib_1.__generator(this, function (_a) {
                            return [2, this.engine.close()];
                        }); }); };
                        return [2];
                }
            });
        });
    };
    Sync.prototype.listener = function () { };
    Object.defineProperty(Sync.prototype, "state", {
        get: function () {
            return this.socket.state;
        },
        enumerable: true,
        configurable: true
    });
    Sync.prototype.onConnectionState = function (func) {
        var _this = this;
        var cancel = false;
        (function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var _a, _b, event_1, e_1_1;
            var e_1, _c;
            return tslib_1.__generator(this, function (_d) {
                switch (_d.label) {
                    case 0:
                        _d.trys.push([0, 5, 6, 11]);
                        _a = tslib_1.__asyncValues(this.socket.listener('connect'));
                        _d.label = 1;
                    case 1: return [4, _a.next()];
                    case 2:
                        if (!(_b = _d.sent(), !_b.done)) return [3, 4];
                        event_1 = _b.value;
                        if (cancel) {
                            return [2];
                        }
                        func(this.state);
                        _d.label = 3;
                    case 3: return [3, 1];
                    case 4: return [3, 11];
                    case 5:
                        e_1_1 = _d.sent();
                        e_1 = { error: e_1_1 };
                        return [3, 11];
                    case 6:
                        _d.trys.push([6, , 9, 10]);
                        if (!(_b && !_b.done && (_c = _a.return))) return [3, 8];
                        return [4, _c.call(_a)];
                    case 7:
                        _d.sent();
                        _d.label = 8;
                    case 8: return [3, 10];
                    case 9:
                        if (e_1) throw e_1.error;
                        return [7];
                    case 10: return [7];
                    case 11: return [2];
                }
            });
        }); })();
        (function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var _a, _b, event_2, e_2_1;
            var e_2, _c;
            return tslib_1.__generator(this, function (_d) {
                switch (_d.label) {
                    case 0:
                        _d.trys.push([0, 5, 6, 11]);
                        _a = tslib_1.__asyncValues(this.socket.listener('disconnect'));
                        _d.label = 1;
                    case 1: return [4, _a.next()];
                    case 2:
                        if (!(_b = _d.sent(), !_b.done)) return [3, 4];
                        event_2 = _b.value;
                        if (cancel) {
                            return [2];
                        }
                        func(this.state);
                        _d.label = 3;
                    case 3: return [3, 1];
                    case 4: return [3, 11];
                    case 5:
                        e_2_1 = _d.sent();
                        e_2 = { error: e_2_1 };
                        return [3, 11];
                    case 6:
                        _d.trys.push([6, , 9, 10]);
                        if (!(_b && !_b.done && (_c = _a.return))) return [3, 8];
                        return [4, _c.call(_a)];
                    case 7:
                        _d.sent();
                        _d.label = 8;
                    case 8: return [3, 10];
                    case 9:
                        if (e_2) throw e_2.error;
                        return [7];
                    case 10: return [7];
                    case 11: return [2];
                }
            });
        }); })();
        return function () { return (cancel = true); };
    };
    Sync.prototype.loop = function () {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!(this.state !== 'open')) return [3, 2];
                        return [4, this.socket.listener('connect').once()];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2:
                        if (this.state !== 'open') {
                            this.loop();
                        }
                        else {
                            this.connect();
                        }
                        return [2];
                }
            });
        });
    };
    return Sync;
}());

exports.CHANNELS = CHANNELS;
exports.Sync = Sync;
exports.batchSize = batchSize;
exports.batchTransfer = batchTransfer;
exports.defaultInterval = defaultInterval;
exports.getBigger = getBigger;
exports.getLastItemRev = getLastItemRev;
exports.hashCode = hashCode;
exports.maxRetries = maxRetries;
exports.maxTimeout = maxTimeout;
exports.syncstateTable = syncstateTable;
exports.waitFor = waitFor;
