import * as tslib_1 from "tslib";
import { CHANNELS } from '../types';
import { delta } from './delta';
import { basic } from './basic';
var SyncEngine = (function () {
    function SyncEngine(socket, db, state) {
        this.initialSyncComplete = Promise.resolve();
        this.cancels = [];
        this.tasks = {};
        this.isClosing = false;
        this.socket = socket;
        this.db = db;
        this.syncState = state;
    }
    Object.defineProperty(SyncEngine.prototype, "isInitialComplete", {
        get: function () {
            return this.tasks['initial'] !== true;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SyncEngine.prototype, "isSyncing", {
        get: function () {
            return Object.keys(this.tasks).length > 0;
        },
        enumerable: true,
        configurable: true
    });
    SyncEngine.prototype.initial = function (collections, isDelta) {
        var release = this.registerTask('initial');
        this.initialSyncComplete = this.initialSync(collections, isDelta).finally(function () {
            release();
        });
        this.cancels.push(this.listenToDatabase(collections, isDelta));
        this.listenToSync();
    };
    SyncEngine.prototype.close = function () {
        var _this = this;
        this.isClosing = true;
        this['_close'] = function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4, Promise.all(this.cancels.map(function (x) { return x(); }))];
                    case 1:
                        _a.sent();
                        return [2];
                }
            });
        }); };
    };
    SyncEngine.prototype.listenToSync = function () {
        var cancels = [];
        cancels.push(delta.listen(this));
        cancels.push(basic.listen(this));
    };
    SyncEngine.prototype.getRemoteChangeCount = function (socket, collections, syncState) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var remoteCounts;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        remoteCounts = {};
                        return [4, Promise.all(collections.map(function (key) {
                                return socket
                                    .invoke(CHANNELS.COUNT_INITIAL_DELTA, {
                                    type: key,
                                    since: syncState.remote(key)
                                })
                                    .then(function (result) { return (remoteCounts[key] = result); });
                            }))];
                    case 1:
                        _a.sent();
                        return [2, remoteCounts];
                }
            });
        });
    };
    SyncEngine.prototype.getLocalChangeCount = function (db, collections, syncState) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var localCounts;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        localCounts = {};
                        return [4, Promise.all(collections.map(function (key) {
                                var begin = syncState.local(key);
                                return db
                                    .count(key, {
                                    where: begin ? ['rev > ?', begin] : undefined
                                })
                                    .then(function (result) { return (localCounts[key] = result); });
                            }))];
                    case 1:
                        _a.sent();
                        return [2, localCounts];
                }
            });
        });
    };
    SyncEngine.prototype.getTaskKey = function () {
        var key = new Date().getTime();
        var count = 0;
        while (this.tasks[key + "-" + count] !== undefined) {
            count = count + 1;
        }
        return key + "-" + count;
    };
    SyncEngine.prototype.registerTask = function (initialKey) {
        var _this = this;
        var key = initialKey || this.getTaskKey();
        this.tasks[key] = true;
        return function () {
            delete _this.tasks[key];
        };
    };
    SyncEngine.prototype.listenToDatabase = function (collections, isDelta) {
        var _this = this;
        return this.db.listen('*', function (error, items, options, key) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var release;
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (error) {
                            throw error;
                        }
                        return [4, this.initialSyncComplete];
                    case 1:
                        _a.sent();
                        if (this.socket.state !== 'open') {
                            return [2];
                        }
                        release = this.registerTask();
                        if (!key ||
                            collections.indexOf(key) === -1 ||
                            options.synced === this.socket.id) {
                            return [2];
                        }
                        if (isDelta(key)) {
                            delta.up(key, items, options, this);
                        }
                        else {
                            basic
                                .up(key, items, options, this)
                                .catch(function (err) { return console.log('ERR', err, err.stack, _this.socket.state); })
                                .finally(function () {
                                release();
                            });
                        }
                        return [2];
                }
            });
        }); });
    };
    SyncEngine.prototype.initialSync = function (collections, isDelta) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var remoteCounts, localCounts, collections_1, collections_1_1, key, down, persist, up, _a, local2, remote2, _b, local, remote, e_1_1;
            var e_1, _c;
            return tslib_1.__generator(this, function (_d) {
                switch (_d.label) {
                    case 0: return [4, this.getRemoteChangeCount(this.socket, collections, this.syncState)];
                    case 1:
                        remoteCounts = _d.sent();
                        return [4, this.getLocalChangeCount(this.db, collections, this.syncState)];
                    case 2:
                        localCounts = _d.sent();
                        _d.label = 3;
                    case 3:
                        _d.trys.push([3, 10, 11, 12]);
                        collections_1 = tslib_1.__values(collections), collections_1_1 = collections_1.next();
                        _d.label = 4;
                    case 4:
                        if (!!collections_1_1.done) return [3, 9];
                        key = collections_1_1.value;
                        down = isDelta(key) ? delta.initialDown : basic.initialDown;
                        return [4, down(key, this.syncState.remote(key), remoteCounts[key], this)];
                    case 5:
                        persist = _d.sent();
                        up = isDelta(key) ? delta.initialUp : basic.initialUp;
                        return [4, up(key, this.syncState.local(key), localCounts[key], this)];
                    case 6:
                        _a = tslib_1.__read.apply(void 0, [_d.sent(), 2]), local2 = _a[0], remote2 = _a[1];
                        return [4, persist()];
                    case 7:
                        _b = tslib_1.__read.apply(void 0, [_d.sent(), 2]), local = _b[0], remote = _b[1];
                        this.syncState.update(key, local, remote);
                        this.syncState.update(key, local2, remote2);
                        _d.label = 8;
                    case 8:
                        collections_1_1 = collections_1.next();
                        return [3, 4];
                    case 9: return [3, 12];
                    case 10:
                        e_1_1 = _d.sent();
                        e_1 = { error: e_1_1 };
                        return [3, 12];
                    case 11:
                        try {
                            if (collections_1_1 && !collections_1_1.done && (_c = collections_1.return)) _c.call(collections_1);
                        }
                        finally { if (e_1) throw e_1.error; }
                        return [7];
                    case 12: return [2];
                }
            });
        });
    };
    return SyncEngine;
}());
export { SyncEngine };
//# sourceMappingURL=index.js.map